import base64
import grp
import os
import pwd
import shutil
import subprocess
import time

import yaml

from charmhelpers import fetch
from charmhelpers.core import hookenv
from charmhelpers.core import host
from charmhelpers.core.services.base import (
    service_restart,
    service_stop,
    )
from charmhelpers.payload import (
    archive,
    execd,
    )


hooks = hookenv.Hooks()
config = hookenv.config()

# Globals
CHARM_FILES_DIR = os.path.join(hookenv.charm_dir(), 'files')
REQUIRED_PACKAGES = [
    'python-virtualenv', 'python-dev', 'python-pygit2', 'git', 'cgit',
    # Unfortunately we need build-essential to compile some extensions,
    # notably Twisted.  Using wheels rather than a pip-cache might let us
    # avoid this in future.
    'build-essential',
    ]
BASE_DIR = config['base_dir']
CODE_DIR = os.path.join(BASE_DIR, 'code')
VENV_DIR = os.path.join(BASE_DIR, 'venv')
LOGS_DIR = os.path.join(BASE_DIR, 'logs')
DATA_DIR = os.path.join(BASE_DIR, 'data')
KEY_DIR = os.path.join(BASE_DIR, 'keys')

CODE_USER = config['code_user']
CODE_GROUP = config['code_group']
USER = config['user']
GROUP = config['group']
CGIT_USER = config['cgit_user']
CGIT_GROUP = config['cgit_group']
PRIVATE_KEY = config['private_ssh_key']
PRIVATE_KEY_PATH = (os.path.join(KEY_DIR, 'ssh-host-key'))
PUBLIC_KEY = config['public_ssh_key']
PUBLIC_KEY_PATH = (os.path.join(KEY_DIR, 'ssh-host-key.pub'))


def get_ports():
    return [s.strip() for s in config['ports'].split(',')]  # no array type


# Environment
def make_srv_location():
    hookenv.log('Creating directories...')

    for dir in (BASE_DIR, CODE_DIR):
        host.mkdir(dir, owner=CODE_USER, group=CODE_GROUP, perms=0o755)
    for dir in (LOGS_DIR, DATA_DIR, KEY_DIR):
        host.mkdir(dir, owner=USER, group=GROUP, perms=0o755)


def execd_preinstall(service_name):
    execd.execd_preinstall()


def create_users(service_name):
    hookenv.log('Creating users and groups for %s.' % service_name)
    code_password = host.pwgen()
    host.adduser(CODE_USER, code_password)
    host.add_user_to_group(CODE_USER, CODE_GROUP)
    password = host.pwgen()
    host.adduser(USER, password)
    host.add_user_to_group(USER, GROUP)
    cgit_password = host.pwgen()
    host.adduser(CGIT_USER, cgit_password)
    host.add_user_to_group(CGIT_USER, CGIT_GROUP)
    # This lets the main service user execute cgitwrap.
    host.add_user_to_group(USER, CGIT_GROUP)


def unpack_source(service_name):
    hookenv.log('Deploying source...')

    make_srv_location()

    # Copy source archive
    archive_path = os.path.join(BASE_DIR, 'turnip.tar.gz')

    with open(os.path.join(CHARM_FILES_DIR, 'turnip.tar.gz')) as file:
        host.write_file(archive_path, file.read(), perms=0o644)

    # Unpack source
    archive.extract_tarfile(archive_path, CODE_DIR)
    os.chown(
        CODE_DIR,
        pwd.getpwnam(CODE_USER).pw_uid, grp.getgrnam(CODE_GROUP).gr_gid)
    host.lchownr(CODE_DIR, CODE_USER, CODE_GROUP)


def install_packages(service_name):
    hookenv.log('Installing system packages...')
    fetch.configure_sources(update=True)
    fetch.apt_install(REQUIRED_PACKAGES, fatal=True)


def install_python_packages(service_name):
    hookenv.log('Installing Python dependencies...')
    pip_cache = os.path.join(CHARM_FILES_DIR, 'pip-cache')
    code_reqs = os.path.join(CODE_DIR, 'requirements.txt')
    deploy_reqs = os.path.join(hookenv.charm_dir(), 'deploy-requirements.txt')

    pip_bin = os.path.join(VENV_DIR, 'bin', 'pip')

    subprocess.call([
        'sudo', '-u', CODE_USER, 'virtualenv', '--system-site-packages',
        VENV_DIR])
    subprocess.check_call([
        'sudo', '-u', CODE_USER, pip_bin, 'install', '--no-index',
        '--find-links={}'.format(pip_cache), '-r', code_reqs,
        '-r', deploy_reqs])
    subprocess.check_call([
        'sudo', '-u', CODE_USER, pip_bin, 'install', '--no-deps',
        '-e', CODE_DIR])


def write_ssh_keys(service_name):
    if PUBLIC_KEY and PRIVATE_KEY:
        keys = [
            {'type': 'public', 'key': PUBLIC_KEY, 'path': PUBLIC_KEY_PATH},
            {'type': 'private', 'key': PRIVATE_KEY, 'path': PRIVATE_KEY_PATH}
            ]
        for key in keys:
            hookenv.log(
                "Writing {} ssh key from config to: {}".format(
                    key['type'], key['path']))
            with open(key['path'], 'w') as f:
                f.write(str(base64.b64decode(key['key'])))
                config['{}_ssh_key_path'.format(key['type'])] = key['path']
                config.save()


def ensure_repo_store_writable(service_name):
    os.chown(DATA_DIR, pwd.getpwnam(USER).pw_uid, grp.getgrnam(GROUP).gr_gid)
    os.chmod(DATA_DIR, 0o755)
    repo_store = os.path.join(DATA_DIR, 'repos')
    if not os.path.exists(repo_store):
        os.makedirs(repo_store)
    os.chown(repo_store, pwd.getpwnam(USER).pw_uid, grp.getgrnam(GROUP).gr_gid)
    os.chmod(repo_store, 0o755)


def publish_wsgi_relations(service_name):
    # Publish the wsgi-file relation so the gunicorn subordinate can
    # serve us. Other WSGI containers could be made to work, as the most
    # gunicorn-specific thing is the --paste hack.
    config = hookenv.config()
    env = {
        'PATH': '%s:$PATH' % os.path.join(VENV_DIR, 'bin'),
        'REPO_STORE': os.path.join(DATA_DIR, 'repos'),
        'TURNIP_LOG_DIR': LOGS_DIR,
        'VIRTINFO_ENDPOINT': config['virtinfo_endpoint'],
        'AUTHENTICATION_ENDPOINT': config['authentication_endpoint'],
        }
    for relid in hookenv.relation_ids('wsgi-file'):
        hookenv.relation_set(
            relid,
            working_dir=CODE_DIR,
            wsgi_wsgi_file='--paste api.ini',  # XXX: Gross.
            wsgi_user=config['user'],
            wsgi_group=config['group'],
            wsgi_log_file=os.path.join(LOGS_DIR, 'api.log'),
            wsgi_timestamp=time.time(),
            port=config['port_api'],
            env_extra=' '.join('%s=%s' % (k, v) for (k, v) in env.items()))


def publish_website_relations(service_name):
    # Due to https://launchpad.net/bugs/1409763, this is an action rather
    # than a provided_data item.
    #
    # haproxy:reverseproxy can only relate to http interfaces.  Rather than
    # providing things like git-service-smart-ssh on nonsensical http
    # interfaces, we instead provide an artificial "website" relation that
    # tells haproxy about everything.
    config = hookenv.config()
    server_name = hookenv.local_unit().replace('/', '-')
    server_ip = str(hookenv.unit_private_ip())
    tcp_options = ['mode tcp', 'balance leastconn', 'option tcplog']
    http_options = ['mode http', 'balance leastconn', 'option httplog']
    haproxy_services = yaml.dump([
        {
            'service_name': 'git-service-git',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_pack_frontend']),
            'service_options': tcp_options,
            'servers': [[
                server_name, server_ip, str(config['port_pack_frontend']),
                'check',
                ]],
            },
        # This is a slightly dubious use of the haproxy charm, but we don't
        # have good relation settings and end up needing to hardcode some
        # ports.
        {
            'service_name': 'git-service-http-redirect',
            'service_host': '0.0.0.0',
            'service_port': '80',
            'service_options':
                http_options +
                ['redirect scheme https code 301 if !{ ssl_fc }'],
            'servers': [],
            },
        {
            'service_name': 'git-service-smart-http',
            'service_host': '0.0.0.0',
            'service_port': '443',
            'service_options':
                http_options +
                ['option httpchk',
                 'rspadd Strict-Transport-Security:\\ max-age=15768000',
                 ],
            'crts': ['DEFAULT'],
            'servers': [[
                server_name, server_ip, str(config['port_smart_http']),
                'check',
                ]],
            },
        {
            'service_name': 'git-service-smart-ssh',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_smart_ssh']),
            'service_options': tcp_options,
            'servers': [[
                server_name, server_ip, str(config['port_smart_ssh']),
                'check',
                ]],
            },
        {
            'service_name': 'turnip-api',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_api']),
            'service_options': http_options + ['option httpchk /repo'],
            'servers': [[
                server_name, server_ip, str(config['port_api']),
                'check',
                ]],
            },
        ])
    for relid in hookenv.relation_ids('website'):
        hookenv.relation_set(
            relid,
            hostname=hookenv.unit_private_ip(),
            port='443',
            services=haproxy_services)


def turnip_service_stop(service_name):
    service_stop('turnip-httpserver')
    service_stop('turnip-sshserver')
    service_stop('turnip-packfrontendserver')
    service_stop('turnip-virtserver')
    service_stop('turnip-packbackendserver')


def turnip_service_restart(service_name):
    service_restart('turnip-packbackendserver')
    service_restart('turnip-virtserver')
    service_restart('turnip-packfrontendserver')
    service_restart('turnip-sshserver')
    service_restart('turnip-httpserver')


def install_nrpe_scripts(service_name):
    src = os.path.join(CHARM_FILES_DIR, "nrpe")
    dst = "/usr/local/lib/nagios/plugins"
    if not os.path.exists(dst):
        os.makedirs(dst)
    shutil.copy2(
        os.path.join(src, "check_git_refs"),
        os.path.join(dst, "check_git_refs"))
    nagios_logs_dir = "/var/log/nagios"
    try:
        nagios_uid = pwd.getpwnam('nagios').pw_uid
        nagios_gid = grp.getgrnam('nagios').gr_gid
        if not os.path.exists(nagios_logs_dir):
            os.mkdir(nagios_logs_dir)
            os.chown(nagios_logs_dir, nagios_uid, nagios_gid)
    except Exception:
        hookenv.log(
            "nagios user not set up; {} not created".format(nagios_logs_dir))


def nagios_restart(service_name):
    service_restart('nagios-nrpe-server')
